/*
 * cli_boot.c   DHCPv6 Client Reboot Program
 *
 * Version:     @(#)cli_boot.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include <db.h>
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dispatch.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_adm.h"
#include "dh6_db.h"
#include "parse.h"
#include "dh6_if.h"
#include "dh6_cli.h"

extern UNION_LIST	server_list,
		 	relay_list,
		 	client_list;
extern IF_INFO 		*this_ifi;

int read_dh6_client_boot( void )
{
	DB  *db;
	DBT key, rec;
	struct cli_ifaddr *cia;

        if ( (db=dh6db_open(DH6CLI_BOOT)) == NULL)
                log_sys(1501, "dbopen error");
	if ((cia=cia_create()) == NULL)
		return -1;

        memset(&key, 0, sizeof(key));
	key.size = KPREFIX_LEN;
	key.data = &key_prefix;
	db->seq(db, &key, &rec, R_CURSOR);

	if (rec.size < sizeof(struct cli_record)) {
		cia_destroy(cia);
        	dh6db_close(db);
		return 0;
	}

	dh6db_convert_rec(&rec, cia);
	do {
		u_queue_tail(&client_list, cia);
		if ((cia=cia_create()) == NULL)
			return -1;
	} while (dh6db_skip_next(db, cia) == 0);

	cia_destroy(cia);
        dh6db_close(db);

	return 0;
}


/*
 *	Be careful!!! The length of a key should be multiple of 16!
 */
int dh6db_alloc_key(DBT *key, void *entry)
{
	struct cli_ifaddr *cia = entry;
        key->size = 2*IPV6_ADDR_LEN;
        key->data = malloc(key->size);
        dh6db_copy(key->data, ifi_lladdr(cia->cia_ifi), IPV6_ADDR_LEN);
        dh6db_copy(((u_int8_t *)key->data)+IPV6_ADDR_LEN, 
		&cia->cia_rly->cr_addr, IPV6_ADDR_LEN); 
}

int dh6db_alloc_rec(DBT *rec, void *entry)
{
	struct cli_ifaddr *cia = entry;
	struct cli_record *crec;
        struct timeval tv;

	rec->size = sizeof(struct cli_record);
        rec->data = malloc(rec->size);
        crec = rec->data;
	strcpy(crec->ifname, ifi_ifname(cia->cia_ifi));
	dh6db_copy(&crec->svr, &cia->cia_svr->cs_addr, IPV6_ADDR_LEN);
	dh6db_copy(&crec->rly, &cia->cia_rly->cr_addr, IPV6_ADDR_LEN);

	/*
 	 *	Very dependent on the structures
	 *	struct cli_ifaddr and struct cli_record !!!
	 */
	dh6db_copy(&crec->addr, &cia->cia_addr, 
		sizeof(struct cli_record) - 2*IPV6_ADDR_LEN -IFNAMSIZ);
        
        gettimeofday(&tv,NULL);
        crec->req_life = crec->req_life*DH6_LIFETIME_UNIT
                            + tv.tv_sec;
        crec->prefer_life = crec->prefer_life*DH6_LIFETIME_UNIT
                            + tv.tv_sec;
        crec->valid_life = crec->valid_life*DH6_LIFETIME_UNIT
                            + tv.tv_sec;
}

int dh6db_convert_rec(DBT *rec, void *entry)
{
	struct cli_ifaddr *cia = entry;
	struct cli_server *cs = NULL;
	struct cli_relay  *cr = NULL;
        struct cli_record *crec = rec->data;
        struct timeval tv;      
        IF_INFO *ifi;

	if (rec->data == NULL || rec->size == 0) {
		log_warn(1515, "Record is NULL");
		return -1;
	}

        ifi = seek_ifi_byname(crec->ifname, strlen(crec->ifname));
        if (ifi != this_ifi)
        	return 1;
        cia->cia_ifi = this_ifi;

        if ((cs=cs_get(&crec->svr)) == NULL) {
        	if ((cs=cs_create()) == NULL)
                        return -1;
                dh6db_copy(&cs->cs_addr,&crec->svr,IPV6_ADDR_LEN);
                cs->cs_rly = NULL;
                u_queue_head(&server_list, cs);
        } 
        if ((cr=cr_get(&crec->rly)) == NULL) {
        	if ((cr=cr_create()) == NULL)
                        return -1;
                dh6db_copy(&cr->cr_addr,&crec->rly,IPV6_ADDR_LEN);
                cr->cr_issvr = NULL;
                u_queue_head(&relay_list, cr);
        } 

	/*
 	 *	Very dependent on the structures
	 *	struct cli_ifaddr and struct cli_record !!!
	 */
	dh6db_copy(&cia->cia_addr, &crec->addr, 
		sizeof(struct cli_record) - 2*IPV6_ADDR_LEN -IFNAMSIZ);
        
        gettimeofday(&tv,NULL);
        cia->cia_req_life = (cia->cia_req_life - tv.tv_sec) /
                                DH6_LIFETIME_UNIT;
        cia->cia_prefer_life = (cia->cia_prefer_life - tv.tv_sec) /
                                DH6_LIFETIME_UNIT;
        cia->cia_valid_life = (cia->cia_valid_life - tv.tv_sec) /
                                DH6_LIFETIME_UNIT;

	cs->cs_rly = cr;
	if ( cr != NULL && cs != NULL &&
	     !memcmp(&cr->cr_addr, &cs->cs_addr, IPV6_ADDR_LEN) )
		cr->cr_issvr = cs;
}

int cli_print_record(struct cli_record *rec)
{
	log_info("DB record: ");
	log_info("  ifname:%s, server: %s", rec->ifname,
		inet_ntop(AF_INET6, &rec->svr, 16, NULL));
	log_info("  agent:  %s", inet_ntop(AF_INET6, &rec->rly, 16, NULL));
	log_info("  client: %s", inet_ntop(AF_INET6, &rec->addr, 16, NULL));
	log_info("  id:%d, prefix:%d, state:%d, req:%ld, prefer:%ld, valid:%ld",
		rec->id, rec->prfsize, rec->state, rec->req_life,
		rec->prefer_life, rec->valid_life);
}

int cli_dbshow_adm( ADM_MESSAGE *msg, int len )
{
	DB  *db;
	DBT key, rec;
	struct cli_record srec;

	/* at the moment just show all the records */
	log_info("Client DB %s:", DH6CLI_BOOT);
        if ( (db=dh6db_open(DH6CLI_BOOT)) == NULL)
                log_sys(1501, "dbopen error");

        memset(&key, 0, sizeof(key));
	key.size = KPREFIX_LEN;
	key.data = &key_prefix;
	db->seq(db, &key, &rec, R_CURSOR);

	do {
		if (rec.size < sizeof(struct cli_record)) 
			break;
		dh6db_copy(&srec, rec.data, sizeof(struct cli_record));
		cli_print_record(&srec);
	} while (db->seq(db, &key, &rec, R_NEXT) == 0);

        dh6db_close(db);
	return 0;
}
