/*
 * dh6_rly.c    DHCPv6 Relay Agent Daemon
 *
 * Version:     @(#)dh6_rly.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dispatch.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_if.h"
#include "dh6_rly.h"
#include "dh6_adm.h"
#include "sock.h"

void rly_run_unix_server( void * );
void rly_adm_msgrcv( void * );
void rly_dh6_msg_rcv( void * );
void rly_dh6_multicast_rcv( void * );
void rs_advertise_timeout( void * );
 
struct fd_list   multi_sock;    /* multicast socket */
struct fd_list   unix_sock;     /* unix socket */
struct poll_list poll_unix;
 
extern UNION_LIST       ifi_table;
UNION_LIST	 server_list, 
		 relay_list;	
 
int     debug = 0;
int     quit_flag = 0;
 
/*
 *	There is a UDP socket for each relay address
 *	There are one multicast socket;
 */

void main( int argc, char *argv[] )
{
	struct ipv6_mreq mr6;
        int     optval;
        char    op;
	struct rly_ifaddr *ria;

        opterr = 0;
        while ((op = getopt(argc, argv, "d:D:")) != EOF) {
                switch (op) {
#ifdef DH6_DEBUG
                        case 'd':   /* debug */
                        case 'D':   /* debug */
                                debug = atoi(optarg);
                                break;
#endif
                        default:
                }
        }

        if ( daemon_init() < 0 )
                log_fatal(2001, "daemon init fail");

        if (debug > 0)
                openlog ("", LOG_CONS, LOG_DAEMON);
        else    
                openlog (argv[0], 0, LOG_DAEMON);

        u_free_list_init();
        begin_timing();

        init_ifi_table();
        get_ifi_table();
#ifdef DH6_DEBUG
        if ( look_cache() ) {
                log_info("debug = %d", debug);
                print_ifi_table();
        }
#endif  

	dh6_relay_init();
	rly_get_addr_table();
#ifdef DH6_DEBUG
        if ( look_cache() ) 
		rly_print_addr_table();
#endif

        if ( (unix_sock.fd=open_unix_sock(dh6rly_adm_server_name())) < 0 )
                log_fatal(2002, "open unix socket fail");

        init_poll(&poll_unix);
        poll_unix.data = &unix_sock;
        poll_unix.function = rly_run_unix_server;
        add_poll(&poll_unix);

	if ( (multi_sock.fd=rly_udp6_socket(NULL)) < 0)
               	log_fatal(2003, "open UDP6 socket fail");

#ifdef IPV6_TXINFO_WORK
        optval = 1;
        if (setsockopt(multi_sock.fd, IPPROTO_IPV6, IPV6_TXINFO,
			(char *)&optval, sizeof(optval)) < 0)
                log_fatal(2004, "IPV6_TXINFO fail");
#endif

        optval = 1;
        if (setsockopt(multi_sock.fd, IPPROTO_IPV6, IPV6_RXINFO,
			(char *)&optval, sizeof(optval)) < 0)
                log_fatal(2005, "IPV6_RXINFO fail");

	init_fdp(&multi_sock);
	multi_sock.data = &multi_sock;
	multi_sock.function = rly_dh6_multicast_rcv;
	add_fdp(&multi_sock);

	/* open a socket for each relay address */
	for (ria = relay_list.ria_head;
		ria != NULL; ria = ria->next) {

		if ( (ria->ria_sk.fd=rly_udp6_socket(&ria->ria_addr)) < 0)
                	log_fatal(2006, "open UDP6 socket fail");

		memcpy(&mr6.ipv6mr_interface, &ria->ria_addr, 
			sizeof(struct in6_addr)); 
		
		/* join relay multicast group */
		memcpy(&mr6.ipv6mr_multiaddr, dh6sin_relay_group(),
					sizeof(struct in6_addr));
		if (setsockopt(ria->ria_sk.fd, IPPROTO_IPV6, IP_ADD_MEMBERSHIP,
						(caddr_t)&mr6, sizeof(mr6)) < 0)
			log_fatal(2007, "relay IP_ADD_MEMBERSHIP fail");
		
		/* join agent multicast group */
		memcpy(&mr6.ipv6mr_multiaddr, dh6sin_agent_group(),
						sizeof(struct  in6_addr));
		if (setsockopt(ria->ria_sk.fd, IPPROTO_IPV6, IP_ADD_MEMBERSHIP,
						(caddr_t)&mr6, sizeof(mr6)) < 0)
			log_fatal(2008, "agent IP_ADD_MEMBERSHIP fail");

	        if (setsockopt(ria->ria_sk.fd, IPPROTO_IPV6, IP_MULTICAST_IF,
	       	                (char *)&ria->ria_addr, IPV6_ADDR_LEN) < 0)
	       	        log_fatal(2009, "IP_MULTICAST_IF fail");

		
	        init_fdp(&ria->ria_sk);
	        ria->ria_sk.data = ria;
	        ria->ria_sk.function = rly_dh6_msg_rcv;
	        add_fdp(&ria->ria_sk);
	}

	read_dh6_relay_conf();	/* read from "dh6_relay.conf" */

        for (;;) {
                dispatch();
                if ( quit_flag )
                        break;
        }

        end_timing();

        del_poll(&poll_unix);
        close_unix_sock(unix_sock.fd, dh6rly_adm_server_name());
	del_fdp(&multi_sock); close(multi_sock.fd);
	for (ria = relay_list.ria_head;
		ria != NULL; ria = ria->next) {
		if (ria->ria_sk.fd > 0) {
			del_fdp(&ria->ria_sk);
			close(ria->ria_sk.fd);
		}
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("exit from %s", argv[0]);
#endif
        closelog();
	exit(0);
}

static inline struct rly_server * rs_create( void )
{
	struct rly_server *rs;
    	if ( (rs=u_alloc_node()) == NULL )
                return NULL;

	rs->rs_ifi = NULL;
        init_timer(&rs->rs_timer);
        rs->rs_timer.data = (void *)rs;
        rs->rs_timer.function = 
		(void (*)(void *))rs_advertise_timeout;

        return rs;
}

static inline void rs_destroy( struct rly_server *rs )
{
        u_free_node(rs);
}

static struct rly_server * rs_get( struct in6_addr *addr )
{
	struct rly_server *rs;
	for (rs = server_list.rs_head;
		rs != NULL; rs = rs->next) {
		if ( !memcmp(&rs->rs_addr, addr, IPV6_ADDR_LEN) )
			return rs;
	}
	return NULL;
}

static struct rly_ifaddr * ria_create( void )
{
	struct rly_ifaddr *ria;
    	if ( (ria=u_alloc_node()) == NULL )
                return NULL;

	ria->ria_ifi = NULL;
	ria->ria_sk.fd = -1;

        return ria;
}

static inline void ria_destroy( struct rly_ifaddr *ria )
{
	if (ria->ria_sk.fd > 0) {
		del_fdp(&ria->ria_sk);
		close( ria->ria_sk.fd );
	}
        u_free_node(ria);
}

struct rly_ifaddr * ria_get(struct in6_addr *addr)
{
	struct rly_ifaddr *ria;
	for (ria = relay_list.ria_head;
		ria != NULL; ria = ria->next) {
		if ( !memcmp(&ria->ria_addr, addr, IPV6_ADDR_LEN) )
			return ria;
	}
	return NULL;
}

struct rly_ifaddr * rly_get_replay_byifindex(int ifindex)
{
        struct rly_ifaddr       *ria;
        for (ria = relay_list.ria_head;
                ria != NULL; ria = ria->next) {
                if (ifi_ifindex(ria->ria_ifi) == ifindex)
                        return ria;
        }
        return NULL;
}

int dh6_relay_init( void )
{
        server_list.chain.head = NULL;
        server_list.chain.head = NULL;
        relay_list.chain.tail = NULL;
        relay_list.chain.tail = NULL;
}

int rly_get_addr_table( void )
{
        u_char buf[8192]; 
        u_char *buf_end;
        struct ifconf ifc;
        struct ifreq *ifr;
        struct if_info *ifi;
        struct rly_ifaddr *ria;
        struct sockaddr_in6 *sin6;
	int s;
 
	if ((s=socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
		log_sys(2010, "socket");
        ifc.ifc_len = sizeof(buf);
        ifc.ifc_buf = (char *)buf; 
        if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0)
                log_fatal(2011,"SIOCGIFCONF failed");
        buf_end = buf + ifc.ifc_len;
        
        ifr = ifc.ifc_req;
        while ((u_char *)ifr < buf_end) {
                sin6 = (struct sockaddr_in6 *)&ifr->ifr_addr;
                if ( ifr->ifr_addr.sa_family == AF_INET6 &&
                     !IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr) ) {
                        if ( (ria = u_alloc_node()) == NULL )
                                return -1;
                        memset(ria, 0, sizeof(struct rly_ifaddr));
                        ria->ria_ifi = seek_ifi_byname(ifr->ifr_name,
                                                strlen(ifr->ifr_name));
                        memcpy(&ria->ria_addr, &sin6->sin6_addr,
                                sizeof(struct in6_addr));
                        u_queue_tail(&relay_list, ria);
                }
                ifr = (struct ifreq *)((char *)ifr + sizeof(ifr->ifr_name) +
                        (sizeof(ifr->ifr_addr) > ifr->ifr_addr.sa_len ?
                         sizeof(ifr->ifr_addr) : ifr->ifr_addr.sa_len ));
        }
	close(s);
        return 0;
}

#ifdef DH6_DEBUG
int rly_print_addr_table( void )
{
	struct rly_ifaddr *ria;
	int n = 1;

	log_info("Server table:");
        for (ria = relay_list.ria_head;
                ria != NULL; ria = ria->next) {
		log_info("    addr: %s, if: %s", 
				inet_ntop(AF_INET6,&ria->ria_addr,16,NULL),
				ifi_ifname(ria->ria_ifi));
	}
	return 0;
}
#endif

struct rly_ifaddr * rly_get_relay_byifname(char *name, int len)
{
	struct rly_ifaddr 	*ria;
	for (ria = relay_list.ria_head;
		ria != NULL; ria = ria->next) {
		if ( !memcmp(ifi_ifname(ria->ria_ifi), name, len) )
			return ria;
	}
	return NULL;
}

struct rly_ifaddr * rly_get_relay_byifindex(int ifindex)
{
	struct rly_ifaddr 	*ria;
	for (ria = relay_list.ria_head;
		ria != NULL; ria = ria->next) {
		if (ifi_ifindex(ria->ria_ifi) == ifindex)
			return ria;
	}
	return NULL;
}


/*
 *  to receive DHCPv6 messages
 */
void rly_dh6_multicast_rcv( void * data )
{
	struct fd_list *sock = (struct fd_list *)data;
    	u_int8_t buf[MAX_MSG_LEN];
	u_int8_t ctrldata[64];

    	for (;;) {
		struct iovec aiov[1];
		struct msghdr msg;
		struct cmsghdr *cmsg = (struct cmsghdr *)ctrldata;
		DH6_MESSAGE    *dh6msg = (DH6_MESSAGE *)buf;
		struct sockaddr_in6 sin6;
    		int  len;

		msg.msg_name = (caddr_t)&sin6;
		msg.msg_namelen = sizeof(struct sockaddr_in6) ;
		msg.msg_iov = aiov;
		msg.msg_iovlen = 1;
		aiov[0].iov_base = buf;
		aiov[0].iov_len = MAX_MSG_LEN;
		msg.msg_control = ctrldata;
		msg.msg_controllen = sizeof(ctrldata);
		msg.msg_flags = 0;
		len = recvmsg(sock->fd, &msg, 0);

        	/*
         	 *  It is very possible for signal SIGCHLD to interrupt this
         	 *  slow system call recvfrom
         	 */
__ugly:
	        if (len == 0)
	            	return;
	        else if (len < 0) {
	            	if (errno == EINTR)
	                	goto __ugly;
                        if (errno == EWOULDBLOCK) 
                                return;  /* no more messages */
	            	log_sys(2012, "recvfrom");
	        }

	        switch ( dh6msg->msg.type ) {
	            	case DH6_SOLICIT:	  
				rly_solicit_rcv(&msg, len);
				break;
	            	case DH6_ADVERTISE:	  
				rly_advertise_rcv(&msg, len);
				break;
	            	default:
	                	log_warn(2013,"unknown message %u", 
					dh6msg->msg.type);
	        }
	}
	return;
}

/*
 *	On receiving DHC Solicit message
 */
int rly_solicit_rcv(struct msghdr *msg, int len)
{
        struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
        struct in6_pktinfo *pktinfo;
        struct rly_ifaddr  *ria;
	int	 extlen = 0;
	void	*ext = NULL;
	struct in6_addr *svr_addrs;
	int	 svr_cnt = 0;
	struct rly_server *rs;
 
#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("receiving Solicit message from %s",
                        inet_ntop(AF_INET6, &src_addr->sin6_addr, 16, NULL));
#endif

       if ((pktinfo=cmsg_get(msg, IPPROTO_IPV6, IPV6_RXINFO)) == NULL) {
                log_warn(2014, "IPV6_RXINFO not found");
                return -1;
        }
        if ((ria=rly_get_relay_byifindex(pktinfo->ipi6_ifindex)) == NULL) {
                log_warn(2015, "relay address entry cannot be located");
                return -1;      
        }

	if ( !dh6msg->sol.Pbit || server_list.rs_head ==NULL ) {
		/* forward it to server */
		rly_solicit_send(&ria->ria_addr,
			         &src_addr->sin6_addr,
				 dh6msg->sol.Cbit,
				 len - dh6_solicit_msg_len(&dh6msg->sol),
				 solicit_ext(&dh6msg->sol));
		return 0;
	}

	/*
	 *	Now I can glean server addresses from my cache
	 *	and build an Advertise message to the client
	 */
#ifdef TODO
	build_extensions_if_any(&ext, &extlen);
#endif

	for (rs = server_list.rs_head; rs != NULL;
			rs = rs->next) {
		/*
		 * DHCP Advertise messages constructed by DHCP relays
		 * from cached server addresses MUST NOT include a
		 * server address on the same link as the soliciting client.
		 */
		if (rs->rs_ifi == ria->ria_ifi)
			continue;
		svr_cnt++;
	}
	svr_addrs = (struct in6_addr *)malloc(svr_cnt*IPV6_ADDR_LEN);
	if (svr_addrs == NULL)
		return -1;
	for (rs = server_list.rs_head; rs != NULL;
			rs = rs->next) {
		if (rs->rs_ifi == ria->ria_ifi)
			continue;
		memcpy(svr_addrs, &rs->rs_addr, IPV6_ADDR_LEN);
	}
	rly_advertise_send(ria, &src_addr->sin6_addr,
			   svr_cnt, svr_addrs, extlen, ext);

	free(svr_addrs);
	free(ext);
	
	return 0;
}

/*
 *	On receiving DHC Advertise message
 */
int rly_advertise_rcv(struct msghdr *msg, int len)
{
        struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
        struct in6_pktinfo *pktinfo;
        struct rly_ifaddr  *ria;
	struct rly_server  *rs;
        struct in6_addr    *svr_addrs; 
	IF_INFO	*ifi;
	int i;

 
#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("receiving Advertise message from %s",
                        inet_ntop(AF_INET6, &src_addr->sin6_addr, 16, NULL));
#endif
       if ((pktinfo=cmsg_get(msg, IPPROTO_IPV6, IPV6_RXINFO)) == NULL) {
                log_warn(2016, "IPV6_RXINFO not found");
                return -1;
        }
        if ((ifi=seek_ifi_byindex(pktinfo->ipi6_ifindex)) == NULL) {
                log_warn(2017, "interface for receiving advertise not located");
                return -1;      
        }

	if ((ria=ria_get(&dh6msg->adv.AgtAddr)) == NULL) {
		log_warn(2019, "Agent address not one of the interfaces");
		return -1;
	}
	
	/* forward it to client if requested by client */
	if ( dh6msg->adv.Lbit )
		rly_request_fwd(ria, len, dh6msg);

        svr_addrs = get_advertise_server_addr(&dh6msg->adv);
        for (i=0; i<dh6msg->adv.SvrCnt; i++) {
                if ( !(rs=rs_get(svr_addrs)) ) {
                        if ( (rs=rs_create()) == NULL )
                                return -1;
                        u_queue_head(&server_list, rs);
			rs->rs_ifi = ifi;
                        memcpy(&rs->rs_addr, svr_addrs, IPV6_ADDR_LEN);
			rs->rs_lifetime = dh6msg->adv.lifetime;
			start_timer(&rs->rs_timer, 
				rs->rs_lifetime*DH6_SVR_TIMEUNIT);
		} else {
			rs->rs_ifi = ifi;
			rs->rs_lifetime = dh6msg->adv.lifetime;
			restart_timer(&rs->rs_timer, 
				rs->rs_lifetime*DH6_SVR_TIMEUNIT);
		}
                svr_addrs++;
        }
	return 0;
}

/*
 *  to receive DHCPv6 messages
 */
void rly_dh6_msg_rcv( void * data )
{
	struct rly_ifaddr *ria = (struct rly_ifaddr *)data;
    	u_int8_t buf[MAX_MSG_LEN];
	u_int8_t ctrldata[64];

    	for (;;) {
		struct iovec aiov[1];
		struct msghdr msg;
		struct cmsghdr *cmsg = (struct cmsghdr *)ctrldata;
		DH6_MESSAGE    *dh6msg = (DH6_MESSAGE *)buf;
		struct sockaddr_in6 sin6;
    		int  len;

		msg.msg_name = (caddr_t)&sin6;
		msg.msg_namelen = sizeof(struct sockaddr_in6) ;
		msg.msg_iov = aiov;
		msg.msg_iovlen = 1;
		aiov[0].iov_base = buf;
		aiov[0].iov_len = MAX_MSG_LEN;
		msg.msg_control = ctrldata;
		msg.msg_controllen = sizeof(ctrldata);
		msg.msg_flags = 0;
		len = recvmsg(ria->ria_sk.fd, &msg, 0);

        	/*
         	 *  It is very possible for signal SIGCHLD to interrupt this
         	 *  slow system call recvfrom
         	 */
__ugly2:
	        if (len == 0)
	            	return;
	        else if (len < 0) {
	            	if (errno == EINTR)
	                	goto __ugly2;
                        if (errno == EWOULDBLOCK) 
                                return;  /* no more messages */
	            	log_sys(2020, "recvfrom");
	        }

	        switch ( dh6msg->msg.type ) {
	            	case DH6_REQUEST: 	  
				rly_request_rcv(ria, &msg, len);
				break;
	            	case DH6_REPLY: 
				rly_reply_rcv(ria, &msg, len);
				break;
	            	default:
	                	log_warn(2021,"unknown message %u", 
					dh6msg->msg.type);
	        }
	}
	return;
}

/*
 *	On receiving DHC Request message
 */
int rly_request_rcv(struct rly_ifaddr *ria, struct msghdr *msg, int len)
{               
        struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
	struct rly_ifaddr *ria2;
 
#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("receiving Request message from %s",
                        inet_ntop(AF_INET6, &src_addr->sin6_addr, 16, NULL));
#endif
    
        if ( !IN6_IS_ADDR_LINKLOCAL(&src_addr->sin6_addr) ) {
                log_warn(2022, "source address of Request not link-local");
                return -1;
        }
	if ( memcmp(&src_addr->sin6_addr, 
		    get_request_llocal_addr(&dh6msg->req),
		    sizeof(struct in6_addr)) ) {
                log_warn(2023, "lladdr in Request not the source address");
		return -1;
	}
	ria2 = ria_get(get_request_agent_addr(&dh6msg->req));
	if (ria2 == NULL || ria2->ria_ifi != ria->ria_ifi) {
                log_warn(2024, "invalid agent address in Request");
		return -1;
	}
	if ( !dh6msg->req.Sbit ) {
		rly_reply_back(ria, msg, DH6_REPLY_POOR_FORM, 0, NULL); 
		return -1;
	}

	rly_request_fwd(ria2, len, dh6msg);
	/*
	 *	TODO:
	 *	a subsequent ICMP from the server should get
	 *	the relay to send a reply with code 64 to the client.
	 */
	return 0;
}

/*
 *	On receiving DHC Reply message
 */
int rly_reply_rcv(struct rly_ifaddr *ria, struct msghdr *msg, int len)
{               
        struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
 
#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("receiving Reply message from %s",
                        inet_ntop(AF_INET6, &src_addr->sin6_addr, 16, NULL));
#endif
    
	if ( !dh6msg->rpl.Lbit ) {
                log_warn(2025, "Lbit == 0 in Reply");
		return -1;
	}
	if ( !IN6_IS_ADDR_LINKLOCAL(get_reply_llocal_addr(&dh6msg->rpl)) ) {
                log_warn(2026, "lladdr in Reply not link-local");
		return -1;
	}
	rly_request_fwd(ria, len, dh6msg);
	return 0;
}

void rly_run_unix_server( void * data )
{     
        struct fd_list          *p = (struct fd_list *)data, *fdp;
        int                     fd;
        struct sockaddr_un      unix_addr;
        int                     unix_len = sizeof(unix_addr);
 
        for (;;) {
                fd = accept(p->fd, (struct sockaddr *)&unix_addr, &unix_len);
                if (fd < 0) {
                        if ( errno!=EWOULDBLOCK )
                                log_sys(2027, "accept");
                        else
                                return;  /* no more pending connections */
                } else {
                        if ( (fdp=malloc(sizeof(struct fd_list))) == NULL )
                                log_sys(2028, "malloc");
                        init_fdp(fdp);
                        fdp->fd   = fd;
                        fdp->data = fdp;
                        fdp->function = rly_adm_msgrcv;
                        add_fdp(fdp);
                }
        }
}

/*
 *	to receive administration commands
 */
void rly_adm_msgrcv( void * data )
{
        struct fd_list *p = (struct fd_list *)data;
        ADM_MESSAGE  msg;
        DH6_MESSAGE *dh6=(DH6_MESSAGE *)(&msg.un.dh6.msg);
        int len;                        
                                
        for (;;) {          
                len = recv(p->fd, (char*)&msg, MAX_MSG_LEN, 0);
                if ( len==0 )
                        return;
                else if (len < 0) {
                        if (errno == EINTR)
                                continue;
                        else
                                log_sys(2029, "recv");
                }

#ifdef DH6_DEBUG
        	if ( look_xmit() )
                        log_info("receiving a message from dh6rc");
#endif
                if (msg.cmd == ADM_CMD_STOP ) {
                        quit_flag = 1;
                        break;
                }

                if (msg.cmd == ADM_DH6_MSG ) {
                        switch ( dh6->msg.type ) {
                        case DH6_SOLICIT:
                                rly_solicit_adm(&msg, len);
                                break;
                        default:
                                log_warn(1128,"unknown message %u",
                                        dh6->msg.type);
                        }
                }
        }
        del_fdp(p); p->fd=-1;
        close(p->fd);
        free(p);
}

/*
 *      On receiving a Solicit message from another process
 */
int rly_solicit_adm( ADM_MESSAGE *msg, int len )
{
        DH6_MESSAGE *dh6=(DH6_MESSAGE *)(&msg->un.dh6.msg);
	struct rly_ifaddr *ria = ria_get(&msg->un.dh6.laddr);

	if (ria == NULL) {
		log_warn(1129, "Invalid agent address");
		return -1;
	}
	rly_solicit_unicast(ria, &msg->un.dh6.faddr, 0, NULL);
        return 0;
}

void rs_advertise_timeout( void * data )
{
	struct rly_server	*rs = (struct rly_server *)data;
	struct rly_ifaddr	*ria = relay_list.ria_head;

	if (ria == NULL) {
                log_warn(2032, "no relay address on list");
		return;
	}
	rly_solicit_unicast(ria, &rs->rs_addr, 0, NULL);
}

int rly_solicit_send(
	struct in6_addr *rlyaddr,   /* address on the receiving interface */
	struct in6_addr *lladdr,    /* link local address of the client */
	int		 Cbit,	    /* copied from original message */
	int		 extlen,
	void 		*ext)
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;
	struct rly_ifaddr *ria;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, dh6sin_server_group(), IPV6_ADDR_LEN);
    	to6.sin6_port = htons(DH6_AGENT_PORT);

    	bzero((char *)msg, MAX_MSG_LEN);
	msg->sol.type = DH6_SOLICIT;
	msg->sol.Cbit = Cbit;
	msg->sol.Lbit = 1;
	msg->sol.Abit = 1;
	msg->sol.Pbit = 0;

	set_solicit_llocal_addr(&msg->sol, lladdr);
	set_solicit_relay_addr(&msg->sol, rlyaddr);

	len = dh6_solicit_msg_len(&msg->sol);

	if ( extlen != 0 && ext != NULL ) {
		memcpy(buf+len, ext, extlen);
		len += extlen;
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Solicit len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                        ntohs(to6.sin6_port));
#endif 

	/*
	 *	The relay should not send sucha multicast solicitation 
	 *	on the interface from which it received the solicitation
	 *	from the client.
	 * 	See page 23, section 7, draft 8.
	 */
	for (ria = relay_list.ria_head; ria != NULL;
			ria = ria->next) {
		if ( !memcmp(&ria->ria_addr, rlyaddr, IPV6_ADDR_LEN) )
			continue;
_resend_solicit:
		if (sendto(ria->ria_sk.fd, msg, len, 0, 
			(struct sockaddr *)&to6, 
			sizeof(struct sockaddr_in6)) != len) {
		        if ( errno==EINTR )
		                goto _resend_solicit;
		        if ( errno==EINVAL ) {
				log_info("Resending DHC Solicit");
		                goto _resend_solicit;
			}
			log_sys(2033, "sendto");
		}
	}
	return 0;
}

int rly_solicit_unicast(
	struct rly_ifaddr *ria, 	/* relay address entry */
	struct in6_addr   *server, 	/* server address */
	int		   extlen,
	void 		  *ext)
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, server, IPV6_ADDR_LEN);
    	to6.sin6_port = htons(DH6_AGENT_PORT);

    	bzero((char *)msg, MAX_MSG_LEN);
	msg->sol.type = DH6_SOLICIT;
	msg->sol.Cbit = 0;
	msg->sol.Lbit = 0;
	msg->sol.Abit = 1;
	msg->sol.Pbit = 0;

	set_solicit_relay_addr(&msg->sol, &ria->ria_addr);

	len = dh6_solicit_msg_len(&msg->sol);

	if ( extlen != 0 && ext != NULL ) {
		memcpy(buf+len, ext, extlen);
		len += extlen;
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("unicasting DHC Solicit len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                        ntohs(to6.sin6_port));
#endif 

_reunicast_solicit:
	if (sendto(ria->ria_sk.fd, msg, len, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _reunicast_solicit;
	        if ( errno==EINVAL )
	                goto _reunicast_solicit;
		log_sys(2034, "sendto");
	}

	return 0;
}

int rly_advertise_send( 
	struct rly_ifaddr *ria, 	/* relay address entry */
	struct in6_addr	  *lladdr,	/* the link local address of client */
	int		   SvrCnt,	/* number of server */
	struct in6_addr	  *svraddr,	/* a list of server addresses */
	int		   extlen,	/* length of extensions */
	void 		  *ext)		/* extensions */
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, lladdr, sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_CLIENT_PORT);

    	bzero((char *)msg, MAX_MSG_LEN);
	msg->adv.type = DH6_ADVERTISE;
	msg->adv.lifetime = 0;  /* unused */
	msg->adv.Lbit = 0;
	msg->adv.Sbit = 0;
	msg->adv.SvrCnt = SvrCnt;
	memcpy(&msg->adv.AgtAddr, &ria->ria_addr, IPV6_ADDR_LEN);
	set_advertise_server_addr(&msg->adv, svraddr, SvrCnt);

	len = dh6_advertise_msg_len(&msg->adv);

	if ( extlen != 0 && ext != NULL ) {
		memcpy(buf+len, ext, extlen);
		len += extlen;
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Advertise len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_advertise:
	if (sendto(ria->ria_sk.fd, msg, len, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_advertise;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Advertise");
	                goto _resend_advertise;
		}
		log_sys(2035, "fd(%d):sendto", ria->ria_sk.fd);
	}
	return 0;
}

int rly_advertise_fwd(
	struct rly_ifaddr *ria,		/* relay entry */
	int		   msglen,
	DH6_MESSAGE	  *msg)
{
	struct sockaddr_in6  to6;

	if ( !msg->adv.Lbit )
		return -1;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, get_advertise_llocal_addr(&msg->adv), 
				sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_CLIENT_PORT);

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("Forwarding Advertise len:%d, to:%s, port:%d",msglen,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_refwd_advertise:
	if (sendto(ria->ria_sk.fd, msg, msglen, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != msglen) {
	        if ( errno==EINTR )
	                goto _refwd_advertise;
	        if ( errno==EINVAL )
	                goto _refwd_advertise;
		log_sys(2036, "sendto");
	}

	return 0;
}

int rly_request_fwd(
	struct rly_ifaddr *ria,		/* relay entry */
	int		   msglen,
	DH6_MESSAGE	  *msg)
{
	struct sockaddr_in6  to6;

	if ( !msg->req.Sbit )
		return -1;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, get_request_server_addr(&msg->req), 
				sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_AGENT_PORT);

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("Forwarding Request len:%d, to: %s, port:%d", msglen,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_request:
	if (sendto(ria->ria_sk.fd, msg, msglen, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != msglen) {
	        if ( errno==EINTR )
	                goto _resend_request;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Request");
	                goto _resend_request;
		}
		log_sys(2037, "sendto");
	}

	return 0;
}

int rly_reply_fwd( 
	struct rly_ifaddr *ria,		/* relay entry */
	int		   msglen,
	DH6_MESSAGE	  *msg)
{
	struct sockaddr_in6  to6;

	if ( !msg->rpl.Lbit )
		return -1;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, get_reply_llocal_addr(&msg->rpl), 
				sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_CLIENT_PORT);

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("Forwarding DHC Reply len:%d, to:%s, port:%d",msglen,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_refwd_reply:
	if (sendto(ria->ria_sk.fd, msg, msglen, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != msglen) {
	        if ( errno==EINTR )
	                goto _refwd_reply;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Reply");
	                goto _refwd_reply;
		}
		log_sys(2038, "fd(%d):sendto", ria->ria_sk.fd);
	}
	return 0;
}

int rly_reply_back( 
	struct rly_ifaddr *ria,		/* relay entry */
        struct msghdr 	  *msg0,	/* original message */
	int     	   code,	/* error code */
	int		   extlen,	/* length of extensions */
	void 		  *ext)		/* extensions */
{
        struct sockaddr_in6 *sdest = (struct sockaddr_in6 *)msg0->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg0->msg_iov[0].iov_base;
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, &sdest->sin6_addr, IPV6_ADDR_LEN);
    	to6.sin6_port = htons(DH6_CLIENT_PORT);

    	bzero((char *)msg, MAX_MSG_LEN);
	msg->rpl.type = DH6_REPLY;
	msg->rpl.code = code;
	msg->rpl.id   = dh6msg->req.id;
	msg->rpl.Lbit = 0;

	len = sizeof(struct dh6_reply);

	if ( extlen != 0 && ext != NULL ) {
		memcpy(reply_ext(&msg->rpl), ext, extlen);
		len += extlen;
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending back DHC Reply len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_reply:
	if (sendto(ria->ria_sk.fd, msg, len, 0, 
		(struct sockaddr *)&to6, sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_reply;
	        if ( errno==EINVAL ) 
	                goto _resend_reply;
		log_sys(2039, "fd(%d):sendto", ria->ria_sk.fd);
	}
	return 0;
}

