/*
 * dispatch.c   Dispatch Module
 *
 * Version:     @(#)dispatch.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include	<stdio.h>
#include	<errno.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	<signal.h>
#include	"u_timer.h"
#include	"dispatch.h"

extern int debug;

static int fd_list_lock_flag = 0;

inline void lock_fd_list( void )
{
	fd_list_lock_flag = 1;
}

inline void unlock_fd_list( void )
{
	fd_list_lock_flag = 0;
}

inline int fd_list_locked( void )
{
	return fd_list_lock_flag;
}

static struct fd_list fdp_head = 
	{ &fdp_head, &fdp_head, FD_SETSIZE, NULL, NULL };
static int nfds = 0;

int reset_fd_list( void )
{
	fdp_head.next = &fdp_head;
	fdp_head.prev = &fdp_head;
	fdp_head.fd   = FD_SETSIZE;
	fdp_head.data = NULL;
	fdp_head.function = NULL;
}

void init_fdp(struct fd_list * fdp)
{
        fdp->next = NULL;
        fdp->prev = NULL;
}

static void reset_nfds( void )
{
	if (fdp_head.prev == &fdp_head)
		nfds = 0;
	else
		nfds = fdp_head.prev->fd + 1;
}

void add_fdp(struct fd_list * fdp)
{
	struct fd_list *p;

	p = &fdp_head;
	do {
		p = p->next;
	} while (fdp->fd > p->fd);

	lock_fd_list();
	fdp->next = p;
	fdp->prev = p->prev;
	p->prev = fdp;
	fdp->prev->next = fdp;
	unlock_fd_list();

	reset_nfds();
}

void del_fdp(struct fd_list * fdp)
{
	if (fdp->next) {
		struct fd_list * next;
		if ((next = fdp->next) != NULL) {
			lock_fd_list();
			(next->prev = fdp->prev)->next = next;
			fdp->next = fdp->prev = NULL;
			reset_nfds();
			unlock_fd_list();
			return;
		}
	}
	return;
}

int dispatch( void )
{
	fd_set 	readfds;
	int	rtn = 0;
	struct fd_list *p, *p2;

	if ( fd_list_locked() )
		return;

	FD_ZERO(&readfds);
	p = fdp_head.next;
	while (p->fd > 0 && p->fd < FD_SETSIZE) {
		FD_SET(p->fd, &readfds);
		p = p->next;
	}

	/* To allow timer interrupt */
        if ( (rtn=select(nfds, &readfds, NULL, NULL, NULL)) < 0) {
                if (errno == EINTR) {
			signal(SIGALRM, run_timer_table);
                	return -1;
		}
                else
                	log_sys_warn(101,"select");
        }
	if (rtn == 0)
		return 0;

	p = fdp_head.next;
	while (p->fd > 0 && p->fd < FD_SETSIZE) {
		void (*fn)(void *) = p->function;
#ifdef DH6_DEBUG
        	if ( look_dp_more() )
			log_info("pid(%d)-dispatch(): fd=%d", getpid(), p->fd);
#endif
		p2 = p; /* in case fn() may remove the current fdp for the queue */
		p = p->next;
                if (FD_ISSET(p2->fd, &readfds)) {
			void *data = p2->data;
			fn(data);
		}
	}
	return rtn;
}

void clean_fd_list( void )
{
	struct fd_list *p, *p2;
	p = fdp_head.next;
	while (p->fd >= 0 && p->fd < FD_SETSIZE) {
		p2 = p;
		p = p->next;
		del_fdp( p2 );
		close( p2->fd );
	}
}
