/*
 * ext6.c    	Extensions Structures for DHCPv6
 *
 * Version:     @(#)ext6.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#ifndef __EXT6_H
#define __EXT6_H

/*
 *	DHCPv6 Pad Extension
 */

struct ext6_pad {
#define EXT6_PAD 	0
	u_int8_t	code;
};

/*
 *	DHCPv6 End Extension
 */

struct ext6_end {
#define EXT6_END 	65535
	u_int16_t	code;
};

/*
 *	DHCPv6 Address Extension
 */

struct ext6_addr {
#define EXT6_ADDR	1
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	rsvd1:3,	
			Pbit:1,		/* P for PTR record ??? */
			Abit:1,		/* A for AAAA record ??? */
			Qbit:1,		/* Q for required by client */
			Lbit:1,		/* L for preferred and valid Lifetime */
			Cbit:1;		/* C for Client address */
	u_int8_t	rsvd2;
	u_int8_t	rsvd3;
	u_int8_t	PrfSize;	/* Prefix size for client address */
	u_int8_t	data[0];
};

/* length of address extension */
extern int ext6_addr_len(struct ext6_addr *);

/* client address if present */
extern struct in6_addr * get_ext6_addr_client_addr(struct ext6_addr *);
extern void set_ext6_addr_client_addr(struct ext6_addr *, struct in6_addr *);

/* preferred lifetime if present */
extern u_int32_t get_ext6_addr_prefer_lifetime(struct ext6_addr *);
extern void set_ext6_addr_prefer_lifetime(struct ext6_addr *, u_int32_t);

/* valid lifetime if present */
extern u_int32_t get_ext6_addr_valid_lifetime(struct ext6_addr *);
extern void set_ext6_addr_valid_lifetime(struct ext6_addr *, u_int32_t);

/* DNS name if present */
extern char * get_ext6_addr_dns_name(struct ext6_addr *);
extern void set_ext6_addr_dns_name(struct ext6_addr *, char *);

/*
 *	DHCPv6 Time Offset Extension
 */

struct ext6_time_offset {
#define EXT6_TIMEOFFSET	8
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	n1;			/* time offset */
	u_int8_t	n2;			/* time offset */
	u_int8_t	n3;			/* time offset */
	u_int8_t	n4;			/* time offset */
};

/*
 *	DHCPv6 Domain Name Server Extension
 */

struct ext6_dns {
#define EXT6_DNS 	9
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	a[1][16];	/* addresses, could be multiple 16 */
};

/*
 *	DHCPv6 Domain Name Extension
 */

struct ext6_dom_name {
#define EXT6_DOM_NAME 	10 
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	d[1];
};

/*
 *	DHCPv6 Directory Agent Extension
 */

struct ext6_da {
#define EXT6_DA       	16
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	ScopeCnt;	/* scope count */
	u_int8_t	rsvd1:6,	
			Mbit:1,		/* M for DA address used only */
			Dbit:1;		/* D for Directory agent address */
	u_int16_t	rsvd2;
	u_int8_t	data[0];
};

/* directory agent address if present */
extern struct in6_addr * get_ext6_da_addr(struct ext6_da *);
extern void set_ext6_da_addr(struct ext6_da *, struct in6_addr *);

/* Scope list: if there is nothing in Scope list, does this list include only '\0'? */
extern char ** get_ext6_da_scope_list(struct ext6_da *);
extern void set_ext6_da_scope_list(struct ext6_da *, char *[]);

/*
 *	DHCPv6 Service Scope Extension
 */

struct ext6_sa {
#define EXT6_SA       	17 
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	scope[0];
};

#ifdef DELETED_IN_DRAFT_08

/*
 *	DHCPv6 Naming Authority Extension
 */

struct ext6_name_auth {
#define EXT6_NAME_AUTH	13 
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	auth[0];
};

#endif

/*
 *	DHCPv6 Static Route Extension
 */

struct ext6_static_route {
#define EXT6_STATIC_ROUTE	24
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	a[1][32];	/* addresses, could be multiple of 32 */
};

/* destination addresses */
extern struct in6_addr * get_ext6_static_route_dest(struct ext6_static_route *, int);
extern void set_ext6_static_route_dest(struct ext6_static_route *, int, struct in6_addr *);

/* router addresses */
extern struct in6_addr * get_ext6_static_route_router(struct ext6_static_route *, int);
extern void set_ext6_static_route_router(struct ext6_static_route *, int, struct in6_addr *);

/*
 *	DHCPv6 TCP Default TTL Extension
 */

struct ext6_tcp_ttl {
#define EXT6_TCP_TTL   	32
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	ttl[4];		/* TTL */
};

#ifdef DELETED_IN_DRAFT_08

/*
 *	DHCPv6 TCP Keepalive Interval Extension
 */

struct ext6_tcp_interval {
#define EXT6_TCP_INTVL	38
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	t1;			/* time */
	u_int8_t	t2;			/* time */
	u_int8_t	t3;			/* time */
	u_int8_t	t4;			/* time */
};

#endif

/*
 *	DHCPv6 Vendor Specific Information Extension
 */

struct ext6_vendor_info {
#define EXT6_VENDOR	40
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	info[0];	/* Vendor Specific Information */
};

/*
 *	DHCPv6 Maximum Message Size Extension
 */

struct ext6_max_msg_size {
#define EXT6_MSG_SIZE	64
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
#define DHCPV6_MIN_SIZE	1500
	u_int16_t	l[2];
};

/*
 *	DHCPv6 Class Identifier Extension
 */

struct ext6_class_id {
#define EXT6_CLASS_ID   65
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	i[1];		/* class identifier */
};

/*
 *	DHCPv6 Reconfigure Multicast Address Extension
 */

struct ext6_multicast {
#define EXT6_MULTICAST  66
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	struct in6_addr	addr;		/* multicast address */
};

/*
 *	DHCPv6 Renumber Server Address Extension
 */

struct ext6_renum_server {
#define EXT6_RENUM_SVR	67
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int8_t	a[16];		/* multicast address */
};

/*
 *	DHCPv6 Client-Server Authentication Extension
 */

struct ext6_authen {
#define EXT6_AUTHEN    	80
	u_int16_t	code;		/* extension type code */
	u_int16_t	len;		/* extension length */
	u_int16_t	sp[4];		/* Security Parameter Index */
	u_int32_t	rp[8];		/* relay protect */
	u_int8_t	auth[0];	/* authenticator */
};

extern u_int8_t * ext6_get(int, void *, int);
extern void * append_addr_ext6(void *, int *, struct in6_addr *, int, 
    u_int32_t, u_int32_t, char *, u_int8_t, u_int8_t, u_int8_t);

#define CLI_APPEND_ADDR_EXT6(ext, len, prefer, valid, Qbit) \
	append_addr_ext6(ext, len, NULL, 0, prefer, valid,  \
			 NULL, Qbit, 0, 0)

#define SVR_APPEND_ADDR_EXT6(ext,len,cliaddr,PrfSize,prefer,valid) \
	append_addr_ext6(ext, len, cliaddr, PrfSize, prefer, valid, \
			 NULL, 0, 0, 0)

extern void * svr_append_multicast_ext6( void *, int *, struct in6_addr *);

#endif /* __EXT6_H */
