/*
 * sock.c    	DHCPv6 Socket Routines
 *
 * Version:     @(#)sock.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_if.h"
#include "dh6_adm.h"
#include "sock.h"


/*
 *	How to configure a socket as a multicast?
 */

/*
 *      Open a UDP socket for an IPv6 client address
 */
static int udp6_socket( struct in6_addr *addr6, u_int16_t port )
{
        int s, optval = 1;
        struct sockaddr_in6  sin6;

      	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
		log_sys(601, "socket");
        fcntl(s, F_SETFL, O_NONBLOCK);

        if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
                          (char *)&optval, sizeof(optval)) < 0)
                log_fatal(602, "SO_REUSEADDR fail");

    	bzero((char *)&sin6, sizeof(sin6));
    	/* 
	 *	u_int8_t  sin6_len; 
	 *	u_int32_t sin6_flowinfo;
	 */
    	sin6.sin6_family = AF_INET6;
	if ( addr6 )
    		memcpy(&sin6.sin6_addr, addr6, sizeof(struct in6_addr));
	else
    		memset(&sin6.sin6_addr, 0, sizeof(struct in6_addr));
    	sin6.sin6_port = htons(port);

        if (bind(s, (struct sockaddr *)&sin6,
                sizeof(struct sockaddr_in6))<0)
		log_sys(603, "bind");
        return s;
}

inline int cli_udp6_socket( struct in6_addr *addr6 )
{
	return udp6_socket(addr6, DH6_CLIENT_PORT);
}

inline int rly_udp6_socket( struct in6_addr *addr6 )
{
	return udp6_socket(addr6, DH6_AGENT_PORT);
}

inline int svr_udp6_socket( struct in6_addr *addr6 )
{
	return udp6_socket(addr6, DH6_AGENT_PORT);
}

char * dh6cli_adm_server_name(char *ifname)
{
	static char name[100];
	sprintf(name, "%sdh6_cli_adm.%s", DH6_SERV_PATH, ifname);
	return name;
}

char * dh6svr_adm_server_name( void )
{
	static char name[100];
	sprintf(name, "%sdh6_svr_adm", DH6_SERV_PATH);
	return name;
}

char * dh6rly_adm_server_name( void )
{
	static char name[100];
	sprintf(name, "%sdh6_rly_adm", DH6_SERV_PATH);
	return name;
}

char * dh6cli_adm_name( void )
{
	static char name[100];
	sprintf(name, "%sdh6_cli_adm", DH6_SERV_PATH);
	return name;
}

/* 
 *	get a UNIX socket
 */
int open_unix_sock( char *name )
{
    	struct sockaddr_un dmn_addr;
    	int  s, dmn_len;

   	unlink( name );  /* in case it was left from last time */

    	/*
     	 *  Open a socket (a UNIX domain stream socket).
    	 */
    	if ( (s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
        	log_sys(604, "socket");
    	fcntl( s, F_SETFL, O_NONBLOCK);

    	/*
    	 *  Bind our local address so that the adm_cli can send to us.
    	 */
    	bzero((char *) &dmn_addr, sizeof(dmn_addr));
    	dmn_addr.sun_family = AF_UNIX;
    	strcpy(dmn_addr.sun_path, name);
    	dmn_len = strlen(dmn_addr.sun_path) + sizeof(dmn_addr.sun_family) + 1;

    	if (bind(s, (struct sockaddr *)&dmn_addr, dmn_len) < 0)
        	log_sys(605, "bind");
    	listen(s, 5);	/* at most 5 messages in queue */

	return s;
}

/*
 *  Close a UDP socket
 */
void close_unix_sock( int s, char *name )
{
	if ( s < 0 )
		return;
	close( s );
   	unlink( name );
}

inline int IN6_IS_ADDR_LINKLOCAL(struct in6_addr *sin6)
{
	return ( *((u_int16_t *)sin6) == ntohs(0xFE80) );
}

inline int IN6_IS_MULTICAST(struct in6_addr *in6)
{
	return ((in6->s6_addr32[0] & htonl(0xffc00000)) == htonl(0xfe800000));
}

int print_sock_bind_info(int s)
{
	struct sockaddr_in6 test_sin;
	int test_len;

    	getsockname(s, (struct sockaddr *)&test_sin, &test_len);
	log_info("socket %d bound to address:%s port:%d", s,
	    inet_ntop(AF_INET6, &test_sin.sin6_addr, 16, NULL), 
	    ntohs(test_sin.sin6_port) );
	return 0;
}

struct sockaddr_in6 * udp6_dest_addr(char *which, ...)
{
	va_list         ap;
        va_start(ap, which);
        if ( !strcmp(which, "sock") ) {		/* derived from socket */
		static struct sockaddr_in6 dest;
		int	s, len;
                s = va_arg(ap, int);
		getsockname(s, (struct sockaddr *)&dest, &len);
        	va_end(ap);
		return &dest;
	} else if ( !strcmp(which, "msg") ) { 	/* derived from message */
		struct msghdr  *msg;
		struct cmsghdr *cmsg;
                msg = va_arg(ap, struct msghdr *);
		cmsg = (struct cmsghdr *)msg->msg_control;
	        if (cmsg->cmsg_level != IPPROTO_IP
	            || cmsg->cmsg_type != IP_RECVDSTADDR) {
	                log_warn(606, "control message type error");
        		va_end(ap);
	                return NULL;
	        }
        	va_end(ap);
		return (struct sockaddr_in6 *)CMSG_DATA(cmsg);
        }
	else {
                log_warn(607, "the way of deriving dest_addr not available");
        	va_end(ap);
		return NULL;
	}
}

void * cmsg_get(struct msghdr *msg, int level, int type)
{
	struct cmsghdr *p;

	if (msg->msg_controllen <= 0)
		return NULL;
	for (p = CMSG_FIRSTHDR(msg); p != NULL;
	     p = CMSG_NXTHDR(msg, p)) {
		if (p->cmsg_level == level &&
		    p->cmsg_type  == type)
			return (void *)CMSG_DATA(p);
	}
	return NULL;
}

void dh6_assign_addr6(
    char 		*ifname, 	/* interface name */
    int 		 namelen, 	/* length of interface name */
    struct in6_addr	*addr6,		/* DHCPv6 address */
    int			 prfsize,	/* prefix size */
    int			 status)	/* status of the address */
{
	struct in6_aliasreq addreq6;
	struct sockaddr_in6 *sin6 = &addreq6.ifra6_addr;
	int s;

	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
                log_sys(608, "socket");

	memcpy(addreq6.ifra6_name, ifname, namelen);

	addreq6.ifra6_mask.sin6_len = sizeof(struct sockaddr_in6);
	inet6_mask(prfsize, &(addreq6.ifra6_mask.sin6_addr));

	sin6->sin6_len = sizeof(struct sockaddr_in6);
	sin6->sin6_family = AF_INET6;
	memcpy(&sin6->sin6_addr, addr6, IPV6_ADDR_LEN);
	addreq6.ifra6_status = 0;

	if ( status == IA6S_PREFERRED)
		addreq6.ifra6_status |= IA6S_DEPRECATED;

	if(ioctl(s, SIOCAIFADDR6, (caddr_t)&addreq6) < 0) 
		log_warn(609, "dh6_assign_addr6 SIOCAIFADDR6 fail");
	close(s);
}

void dh6_remove_addr6(
    char 		*ifname, 	/* interface name */
    int 		 namelen, 	/* length of interface name */
    struct in6_addr	*addr6)		/* DHCPv6 address */
{
	struct in6_ifreq ridreq6;
	struct sockaddr_in6 *sin6 = &ridreq6.ifr6_addr;
	int s;

	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
                log_sys(608, "socket");

	memcpy(ridreq6.ifr6_name, ifname, namelen);

	sin6->sin6_len = sizeof(struct sockaddr_in6);
	sin6->sin6_family = AF_INET6;
	memcpy(&sin6->sin6_addr, addr6, IPV6_ADDR_LEN);

	if(ioctl(s, SIOCDIFADDR6, (caddr_t)&ridreq6) < 0) 
		log_warn(609, "dh6_remove_addr6() SIOCDIFADDR6 fail");
	close(s);
}

