/*
 * utils.c    	DHCPv6 Utilities
 *
 * Version:     @(#)utils.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"

char	*pname = NULL;		/* caller can set this from argv[0] */

extern int debug;

/*
 *	Generate a message
 */

void log_info(const char *ctl, ...)
{
	char buf[1000];
	va_list list;
	va_start (list,ctl);
	vsprintf (buf,ctl,list);
	syslog (LOG_INFO,"%s",buf);
	va_end(list);
}

/*
 *	Generate a warning error message without termination
 */
void log_warn(int err, const char *ctl, ...)
{
	char buf[1000];
	va_list list;
	va_start (list,ctl);
	vsprintf (buf,ctl,list);
	syslog (LOG_ERR,"err%d: %s", err, buf);
	va_end(list);
}
/*
 *	Generate a fatal error message and terminate
 */
void log_fatal(int err, const char *ctl, ...)
{
	char 	buf[1000];
	int	errno_save;
	va_list list;

	va_start (list,ctl);
	errno_save = errno;		/* value caller might want printed */
	vsprintf (buf,ctl,list);
	syslog (LOG_ERR,"err%d: %s (serr%d:%s)", 
		err, buf, errno_save, strerror(errno_save));
	va_end(list);
	exit(1);
}

/* 
 *	Fatal error related to a system call.
 * 	Print a message and terminate. 
 */

void log_sys(int err, const char *ctl, ...)
{
	va_list	list;
	int	errno_save;
	char	buf[1000];

	va_start(list, ctl);
	errno_save = errno;		/* value caller might want printed */
	vsprintf(buf, ctl, list);
	syslog (LOG_ERR, "err%d: %s() - %d:%s", 
		err, buf, errno_save, strerror(errno_save));
	va_end(list);
	exit(2);
}

/* 
 *	Fatal error related to a system call.
 * 	Print a message and terminate. 
 */

int log_sys_warn(int err, const char *ctl, ...)
{
	va_list	list;
	int	errno_save;
	char	buf[1000];

	va_start(list, ctl);
	errno_save = errno;		/* value caller might want printed */
	vsprintf(buf, ctl, list);
	syslog (LOG_ERR, "err%d: %s() - %d:%s", 
		err, buf, errno_save, strerror(errno_save));
	va_end(list);
}

int lock_reg(int fd, int cmd, int type, 
		off_t offset, int whence, off_t len)
{
	struct flock	lock;

	lock.l_type = type;	/* F_RDLCK, F_WRLCK, F_UNLCK */
	lock.l_start = offset;	/* byte offset, relative to l_whence */
	lock.l_whence = whence;	/* SEEK_SET, SEEK_CUR, SEEK_END */
	lock.l_len = len;		/* #bytes (0 means to EOF) */

	return( fcntl(fd, cmd, &lock) );
}

int daemon_init( void )
{
	pid_t pid;

	if ( (pid=fork()) < 0 ) {
		log_sys(401, "fork");
		return -1;
	}

	/*
	 *	the top parent process does nothing and exits
	 */
	else if ( pid!=0 )
		exit(0);	

	setsid();
	chdir("/");
	umask(0);
	return 0;
}

int delay(int k)
{
	int i, j;
	for (i = 0; i < k; i++)
	for (j = 0; j < k; i++)
		;
}

int deadloop( void )
{
	for (;;) ;
}

char * time_ntop(int lifetime)
{
	static char buf[20];
	struct timeval tv;
	struct tm *t;
	time_t clock;

	gettimeofday(&tv,NULL);
	clock = tv.tv_sec + lifetime;
	t = gmtime(&clock);
	sprintf(buf, "%02s-%02s-%02s//%02s:%02s:%02s",
		t->tm_year, t->tm_mon, t->tm_mday, 
		t->tm_hour, t->tm_min, t->tm_sec);
	return buf;
}

int time_pton(char *stm)
{
	struct timeval tv;
	time_t clock;
	struct tm t;

	t.tm_year = atoi(stm);
	t.tm_mon  = atoi(stm+3);
	t.tm_mday = atoi(stm+6);
	t.tm_hour = atoi(stm+9);
	t.tm_min  = atoi(stm+12);
	t.tm_sec  = atoi(stm+15);
	clock = mktime(&t);
	gettimeofday(&tv,NULL);
	if (clock > tv.tv_sec)
		return clock - tv.tv_sec;
	else
		return 0;
}

#ifdef DH6_DEBUG

#define DEBUG_LEVEL	6

inline look_xmit( void )
{
	return (debug%DEBUG_LEVEL >= 1);
}

inline look_xmit_more( void )
{
	return (debug == DEBUG_LEVEL);
}

inline look_cache( void )
{
	return (debug%DEBUG_LEVEL >= 2);
}

inline look_cache_more( void )
{
	return (debug == DEBUG_LEVEL+1);
}

inline look_db( void )
{
	return (debug%DEBUG_LEVEL >= 3);
}

inline look_db_more( void )
{
	return (debug == DEBUG_LEVEL+2);
}

inline look_dp( void )
{
	return (debug%DEBUG_LEVEL >= 4);
}

inline look_dp_more( void )
{
	return (debug == DEBUG_LEVEL+3);
}

inline look_tm( void )
{
	return (debug%DEBUG_LEVEL >= 5);
}

inline look_tm_more( void )
{
	return (debug == DEBUG_LEVEL+4);
}

#endif /* DH6_DEBUG */
